"""TotalConnectClient() in this file is the primary class of this package.

Instantiate it like this:

usercodes = { 'default': '1234' }
client = TCC.TotalConnectClient(username, password, usercodes)

for location in client.locations:
    ### do stuff with this location
"""

import base64
import json
import logging
import time
from typing import Any, Callable, Dict

import requests
from Crypto.Cipher import PKCS1_v1_5
from Crypto.PublicKey import RSA
from oauthlib.oauth2 import LegacyApplicationClient, OAuth2Error
from requests_oauthlib import OAuth2Session

from .const import (
    AUTH_CONFIG_ENDPOINT,
    AUTH_TOKEN_ENDPOINT,
    HTTP_API_LOGOUT,
    HTTP_API_SESSION_DETAILS_ENDPOINT,
    ArmType,
    _ResultCode,
)
from .exceptions import (
    AuthenticationError,
    BadResultCodeError,
    FailedToBypassZone,
    FeatureNotSupportedError,
    InvalidSessionError,
    RetryableTotalConnectError,
    ServiceUnavailable,
    TotalConnectError,
    UsercodeInvalid,
    UsercodeUnavailable,
)
from .location import TotalConnectLocation
from .user import TotalConnectUser

DEFAULT_USERCODE = "-1"

LOGGER = logging.getLogger(__name__)


class TotalConnectClient:
    """Client for Total Connect."""

    TIMEOUT = 60  # seconds until I/O will fail
    MAX_RETRY_ATTEMPTS = 5  # number of times to retry an API call

    def __init__(  # pylint: disable=too-many-arguments
        self,
        username: str,
        password: str,
        usercodes: Dict[str, str] | None = None,
        auto_bypass_battery: bool = False,
        retry_delay: int = 6,  # seconds between retries
    ) -> None:
        """Initialize."""
        self.times = {}
        self.time_start = time.time()

        self.username: str = username
        self.password: str = password
        self.usercodes = usercodes or {}
        self.auto_bypass_low_battery = auto_bypass_battery
        self.retry_delay = retry_delay

        self._logged_in = False
        self._oauth_session = None
        self._oauth_client = None
        self._invalid_credentials = False
        self._client_id = None
        self._app_id = None
        self._app_version = None
        self._key_pem = None

        self._module_flags: Dict[str, str] = {}
        self._user: TotalConnectUser | None = None
        self._locations: Dict[Any, TotalConnectLocation] = {}
        self._locations_unfetched: Dict[Any, TotalConnectLocation] = {}

        self.authenticate()

        self.times["__init__"] = time.time() - self.time_start

    @property
    def locations(self) -> Dict[Any, TotalConnectLocation]:
        """Raises an exception if the panel cannot be reached to retrieve
        metadata or details. This can be retried later and will succeed
        if/when the panel becomes reachable.
        """
        # to_fetch is needed because items() is invalidated by del
        to_fetch = list(self._locations_unfetched.items())
        for locationid, location in to_fetch:
            try:
                location.get_partition_details()
                location.get_zone_details()
                location.get_panel_meta_data()
                # if we get here, it has been fetched successfully
                del self._locations_unfetched[locationid]
            except Exception:
                LOGGER.error(f"exception during initial fetch of {locationid}")
                raise
        assert not self._locations_unfetched
        return self._locations

    def __str__(self) -> str:
        """Return a text string that is printable."""
        data = (
            f"CLIENT\n\n"
            f"Username: {self.username}\n"
            f"Password: {'[hidden]' if self.password else '[unset]'}\n"
            f"Usercode: {self.usercodes}\n"
            f"Auto Bypass Low Battery: {self.auto_bypass_low_battery}\n"
            f"Invalid Credentials: {self._invalid_credentials}\n"
            f"Module Flags:\n"
        )

        for key, value in self._module_flags.items():
            data = data + f"  {key}: {value}\n"

        data = data + str(self._user)

        locations = f"LOCATIONS: {len(self.locations)}\n\n"
        for location in self.locations:
            locations = locations + str(self.locations[location])

        return data + locations

    def times_as_string(self) -> str:
        """Return a string with times."""
        self.times["total running time"] = time.time() - self.time_start
        msg = "total-connect-client time info (seconds):\n"
        for key, value in self.times.items():
            msg = msg + f"  {key}: {value}\n"

        return msg

    def _raise_for_retry(self, response: Dict[str, Any]) -> None:
        """Determine which responses should be retried in request()."""
        rc = _ResultCode.from_response(response)
        if rc == _ResultCode.INVALID_SESSION:
            raise InvalidSessionError("invalid session", response)
        if rc == _ResultCode.INVALID_SESSIONID:
            raise InvalidSessionError("invalid session ID", response)
        if rc == _ResultCode.CONNECTION_ERROR:
            raise RetryableTotalConnectError("connection error", response)
        if rc in (_ResultCode.FAILED_TO_CONNECT, _ResultCode.CANNOT_CONNECT):
            raise RetryableTotalConnectError("failed to connect with panel", response)
        if rc == _ResultCode.BAD_OBJECT_REFERENCE:
            raise RetryableTotalConnectError("bad object reference", response)

    def raise_for_resultcode(self, response: Dict[str, Any]) -> None:
        """If response.ResultCode indicates success, return and do nothing.

        If it indicates an authentication error, raise AuthenticationError.
        """
        rc = _ResultCode.from_response(response)
        if rc in (
            _ResultCode.SUCCESS,
            _ResultCode.ARM_SUCCESS,
            _ResultCode.DISARM_SUCCESS,
            _ResultCode.SESSION_INITIATED,
        ):
            return
        self._raise_for_retry(response)
        if rc in (
            _ResultCode.BAD_USER_OR_PASSWORD,
            _ResultCode.AUTHENTICATION_FAILED,
            _ResultCode.ACCOUNT_LOCKED,
        ):
            raise AuthenticationError(rc.name, response)
        if rc == _ResultCode.USER_CODE_UNAVAILABLE:
            raise UsercodeUnavailable(rc.name, response)
        if rc == _ResultCode.USER_CODE_INVALID:
            raise UsercodeInvalid(rc.name, response)
        if rc == _ResultCode.FEATURE_NOT_SUPPORTED:
            raise FeatureNotSupportedError(rc.name, response)
        if rc == _ResultCode.FAILED_TO_BYPASS_ZONE:
            raise FailedToBypassZone(rc.name, response)
        raise BadResultCodeError(rc.name, response)


    def _request_with_retries(
        self,
        do_request: Callable[[], Dict[str, Any]],
        request_description: str,
        attempts_remaining: int = MAX_RETRY_ATTEMPTS,
    ):
        """Call a given request function and handle retries for temporary errors and authentication
        problems."""
        is_first_request = attempts_remaining == self.MAX_RETRY_ATTEMPTS
        attempts_remaining -= 1

        try:
            LOGGER.debug(f"sending API request {request_description}")
            response = do_request()
            self._raise_for_retry(response)
            return response
        # To retry an exception that could be raised during the
        # request, add it to an except block here, depending on what
        # you want to have happen. The first block just retries and
        # logs. The second block causes reauthentication.
        except RetryableTotalConnectError as err:
            if attempts_remaining <= 0:
                raise
            msg = f"{self.username} {request_description} {err.args[0]} on response"
            if is_first_request:
                LOGGER.info(f"{msg}: {attempts_remaining} retries remaining")
            else:
                LOGGER.debug(f"{msg}: {attempts_remaining} retries remaining")
            time.sleep(self.retry_delay)
        except requests.RequestException as err:
            if attempts_remaining <= 0:
                raise ServiceUnavailable(
                    f"Error connecting to Total Connect service: {err}"
                ) from err
            LOGGER.debug(
                f"Error connecting to Total Connect service: {attempts_remaining} retries remaining"
            )
            time.sleep(self.retry_delay)
        except (OAuth2Error, InvalidSessionError) as err:
            if attempts_remaining <= 0:
                raise ServiceUnavailable(
                    f"Invalid Session after multiple retries: {err}"
                ) from err
            LOGGER.info(
                f"reauthenticating {self.username}: {attempts_remaining} retries remaining"
            )
            self.authenticate()

        return self._request_with_retries(
            do_request, request_description, attempts_remaining
        )

    def http_request(
        self,
        endpoint: str,
        method: str,
        params: Dict[str, Any] | None = None,
        data: Dict[str, Any] | None = None,
    ) -> Dict[str, Any]:
        """Send an HTTP request to a Web API endpoint

        method is the HTTP method, e.g. 'GET', 'POST', 'PUT', 'DELETE'
        params is a dictionary defining the query parameters to add to the endpoint URL (usually with GET)
        data is a dictionary defining the query parameter to encode in the request body (usually with POST/PUT)
        """
        LOGGER.debug(
            f"\n----- http_request -----\n\tendpoint: {endpoint}\n\tmethod: {method}\n\tparams: {params}\n\tdata: {data}\n----- end request -----"
        )

        def _do_http_request() -> Dict[str, Any]:
            response = self._oauth_session.request(
                method=method, url=endpoint, params=params, data=data
            )
            LOGGER.debug(
                f"\n----- http response -----\n\tok: {response.ok}\n\tstatus code: {response.status_code}\n\tJSON: {response.json()}\n----- end response -----"
            )
            if not response.ok:
                LOGGER.debug(
                    f"Received HTTP error code {response.status_code} with response:",
                    response.content,
                )
            return response.json()

        args = {**(params or {}), **(data or {})}
        return self._request_with_retries(
            _do_http_request, f"{method} {endpoint} ({args})"
        )

    def _encrypt_credential(self, credential: str) -> str:
        # Load the key from the PEM file
        key = RSA.importKey(self._key_pem)

        # Create a cipher object using PKCS1_OAEP padding
        cipher = PKCS1_v1_5.new(key)

        # Encrypt the message
        encrypted_message = cipher.encrypt(credential.encode())

        # Encode the encrypted message in base64 for safe transmission
        return base64.b64encode(encrypted_message).decode()

    def authenticate(self) -> None:
        """Login to the system.

        Upon success, self._logged_in is True,
        and self._user and self.locations are valid.
        self.locations will not be refreshed if it was non-empty on entry.
        """
        start_time = time.time()
        if self._invalid_credentials:
            raise AuthenticationError(
                f"not authenticating: password already failed for user {self.username}"
            )

        self._get_configuration()
        self._request_token()

        # Retrieve user and location information.
        if not self._locations:
            response = self.http_request(
                endpoint=HTTP_API_SESSION_DETAILS_ENDPOINT,
                method="GET",
                params={"appId": self._app_id, "appVersion": self._app_version},
            )["SessionDetailsResult"]
            self._module_flags = dict(
                x.split("=") for x in response["ModuleFlags"].split(",")
            )
            self._user = TotalConnectUser(response["UserInfo"])
            self._locations_unfetched = self._make_locations(response)
            self._locations = self._locations_unfetched.copy()
            if not self._locations:
                raise TotalConnectError("no locations found", response)
        LOGGER.info(f"{self.username} authenticated: {len(self._locations)} locations")
        self.times["authenticate"] = time.time() - start_time

    def _get_configuration(self) -> None:
        """Retrieve application configuration for TotalConnect REST API."""
        response = requests.get(AUTH_CONFIG_ENDPOINT, timeout=self.TIMEOUT)
        if not response.ok:
            raise ServiceUnavailable(
                f"Service configuration is not available at {AUTH_CONFIG_ENDPOINT}"
            )
        config = response.json()
        key = config["AppConfig"][0]["tc2APIKey"]
        self._client_id = config["AppConfig"][0]["tc2ClientId"]
        self._app_id = next(
            info for info in config["brandInfo"] if info["BrandName"] == "totalconnect"
        )["AppID"]
        self._app_version = (
            config["RevisionNumber"] + "." + config["version"].split(".")[-1]
        )
        self._key_pem = (
            "-----BEGIN PUBLIC KEY-----\n" + key + "\n-----END PUBLIC KEY-----"
        )

    def _request_token(self) -> None:
        """Request a token using OAuth2."""
        def token_updater(token):
            """Update the token on auto-refresh.
            
            Called following successful token auto-refresh by OAuth2Session.
            """
            self._logged_in = True
            LOGGER.debug("Session token was auto-refreshed")

        self._oauth_client = LegacyApplicationClient(client_id=self._client_id)
        self._oauth_session = OAuth2Session(
            client_id=self._client_id,
            client=self._oauth_client,
            auto_refresh_url=AUTH_TOKEN_ENDPOINT,
            auto_refresh_kwargs={"client_id": self._client_id},
            token_updater=token_updater
        )
        try:
            self._oauth_session.fetch_token(
                token_url=AUTH_TOKEN_ENDPOINT,
                username=self._encrypt_credential(self.username),
                password=self._encrypt_credential(self.password),
                client_id=self._client_id,
            )
        except OAuth2Error as exc:
            try:
                self.raise_for_resultcode(json.loads(exc.json))
            except AuthenticationError:
                self._invalid_credentials = True
                self._logged_in = False
                raise
        self._logged_in = True

    def is_logged_in(self) -> bool:
        """Return true if the client is logged in to Total Connect."""
        return self._logged_in

    def log_out(self) -> None:
        """Upon return, we are logged out.

        Raises TotalConnectError if we still might be logged in.
        """
        if self.is_logged_in():
            response = self.http_request(endpoint=HTTP_API_LOGOUT, method="POST")
            self.raise_for_resultcode(response)
            if response["ResultCode"] != 0:
                raise TotalConnectError(
                    f"Logout failed with response code {response['ResultCode']}: {response['ResultData']}"
                )
            LOGGER.info("Logout Successful")
            self._logged_in = False

    def get_number_locations(self) -> int:
        """Return the number of locations.

        Home Assistant needs a way to force the locations to load
        inside a callable function.
        """
        return len(self.locations)

    def _make_locations(
        self, response: Dict[str, Any]
    ) -> Dict[Any, TotalConnectLocation]:
        """Return a dict mapping LocationID to TotalConnectLocation."""
        start_time = time.time()
        new_locations = {}

        for locationinfo in response.get("Locations") or []:
            location_id = locationinfo["LocationID"]
            location = TotalConnectLocation(locationinfo, self)
            new_locations[location_id] = location

            location.auto_bypass_low_battery = self.auto_bypass_low_battery

            # set the usercode for the location
            usercode = (
                self.usercodes.get(location_id)  # noqa: W504
                or self.usercodes.get(str(location_id))
                or self.usercodes.get("default")
            )
            if usercode:
                location.usercode = usercode
            else:
                LOGGER.warning(f"no usercode for location {location_id}")
                location.usercode = DEFAULT_USERCODE

        self.times["_make_locations"] = time.time() - start_time
        return new_locations


class ArmingHelper:
    """
    For a partition or location, you can call its arm() or disarm() method directly.

    Example: partition.arm(ArmType.AWAY)

    Alternatively, you can use ArmingHelper.
       Example: ArmingHelper(partition).arm_away()
    """

    def __init__(self, partition_or_location) -> None:
        """Initialize ArmingHelper."""
        self.armable = partition_or_location

    def arm_away(self, usercode: str = "") -> None:
        """Arm the system (Away)."""
        self.armable.arm(arm_type=ArmType.AWAY, usercode=usercode)

    def arm_stay(self, usercode: str = "") -> None:
        """Arm the system (Stay)."""
        self.armable.arm(arm_type=ArmType.STAY, usercode=usercode)

    def arm_stay_instant(self, usercode: str = "") -> None:
        """Arm the system (Stay - Instant)."""
        self.armable.arm(arm_type=ArmType.STAY_INSTANT, usercode=usercode)

    def arm_away_instant(self, usercode: str = "") -> None:
        """Arm the system (Away - Instant)."""
        self.armable.arm(arm_type=ArmType.AWAY_INSTANT, usercode=usercode)

    def arm_stay_night(self, usercode: str = "") -> None:
        """Arm the system (Stay - Night)."""
        self.armable.arm(arm_type=ArmType.STAY_NIGHT, usercode=usercode)

    def disarm(self, usercode: str = "") -> None:
        """Disarm the system."""
        self.armable.disarm(usercode=usercode)



"""Test total_connect_client."""

import unittest
from unittest.mock import Mock

import pytest
from const import (
    ZONE_LOW_BATTERY,
    ZONE_STATUS_LYRIC_CONTACT,
    ZONE_STATUS_LYRIC_LOCAL_ALARM,
    ZONE_STATUS_LYRIC_KEYPAD,
    ZONE_STATUS_LYRIC_MOTION,
    ZONE_STATUS_LYRIC_POLICE,
    ZONE_STATUS_LYRIC_TEMP,
    ZS_NORMAL,
)

from total_connect_client.zone import ZoneStatus, ZoneType
from total_connect_client.zone import TotalConnectZone as tcz
from total_connect_client.exceptions import TotalConnectError

ZONE_BYPASSED = {
    "ZoneDescription": "Bypassed",
    "PartitionId": "1",
    "ZoneTypeId": ZoneType.SECURITY,
    "CanBeBypassed": 1,
    "ZoneStatus": ZoneStatus.BYPASSED,
}

ZONE_FAULTED = {
    "ZoneDescription": "Faulted",
    "ZoneID": "1",
    "PartitionId": "1",
    "ZoneTypeId": ZoneType.SECURITY,
    "CanBeBypassed": 1,
    "ZoneStatus": ZoneStatus.FAULT,
}

ZONE_TAMPERED = {
    "ZoneDescription": "Tampered",
    "PartitionId": "1",
    "ZoneTypeId": ZoneType.SECURITY,
    "CanBeBypassed": 1,
    "ZoneStatus": ZoneStatus.TROUBLE,
}

ZONE_BYPASSED_LOW_BATTERY = {
    "ZoneDescription": "Bypassed Low Battery",
    "PartitionId": "1",
    "ZoneTypeId": ZoneType.SECURITY,
    "CanBeBypassed": 1,
    "ZoneStatus": 65,
}

ZONE_TROUBLE_LOW_BATTERY = {
    "ZoneDescription": "Trouble Low Battery",
    "PartitionId": "1",
    "ZoneTypeId": ZoneType.SECURITY,
    "CanBeBypassed": 1,
    "ZoneStatus": 72,
}

ZONE_TRIGGERED = {
    "ZoneDescription": "Triggered",
    "PartitionId": "1",
    "ZoneTypeId": ZoneType.SECURITY,
    "CanBeBypassed": 1,
    "ZoneStatus": ZoneStatus.TRIGGERED,
}

ZONE_BUTTON = {
    "ZoneDescription": "Button",
    "PartitionId": "1",
    "ZoneTypeId": ZoneType.SECURITY,
    "CanBeBypassed": 0,
    "ZoneStatus": ZoneStatus.NORMAL,
}

ZONE_SMOKE = {
    "ZoneDescription": "Smoke",
    "PartitionId": "1",
    "ZoneTypeId": ZoneType.FIRE_SMOKE,
    "CanBeBypassed": 0,
    "ZoneStatus": ZoneStatus.NORMAL,
}

ZONE_GAS = {
    "ZoneDescription": "Gas",
    "PartitionId": "1",
    "ZoneTypeId": ZoneType.CARBON_MONOXIDE,
    "CanBeBypassed": 0,
    "ZoneStatus": ZoneStatus.NORMAL,
}


class TestTotalConnectZone(unittest.TestCase):
    """Test TotalConnectZone."""

    def setUp(self):
        """Test setup."""
        self.zone_normal = tcz(ZS_NORMAL, None)
        self.zone_bypassed = tcz(ZONE_BYPASSED, None)
        self.zone_faulted = tcz(ZONE_FAULTED, None)
        self.zone_tampered = tcz(ZONE_TAMPERED, None)
        self.zone_low_battery = tcz(ZONE_LOW_BATTERY, None)
        self.zone_bypassed_low_battery = tcz(ZONE_BYPASSED_LOW_BATTERY, None)
        self.zone_trouble_low_battery = tcz(ZONE_TROUBLE_LOW_BATTERY, None)
        self.zone_triggered = tcz(ZONE_TRIGGERED, None)
        self.zone_button = tcz(ZONE_BUTTON, None)
        self.zone_smoke = tcz(ZONE_SMOKE, None)
        self.zone_gas = tcz(ZONE_GAS, None)
        self.zone_lyric_contact = tcz(ZONE_STATUS_LYRIC_CONTACT, None)
        self.zone_lyric_motion = tcz(ZONE_STATUS_LYRIC_MOTION, None)
        self.zone_lyric_police = tcz(ZONE_STATUS_LYRIC_POLICE, None)
        self.zone_lyric_temp = tcz(ZONE_STATUS_LYRIC_TEMP, None)
        self.zone_lyric_keypad = tcz(ZONE_STATUS_LYRIC_KEYPAD, None)
        self.zone_lyric_local_alarm = tcz(ZONE_STATUS_LYRIC_LOCAL_ALARM, None)

    def tearDown(self):
        """Tear down."""
        self.zone_normal = None
        self.zone_bypassed = None
        self.zone_faulted = None
        self.zone_tampered = None
        self.zone_low_battery = None
        self.zone_trouble_low_battery = None
        self.zone_triggered = None
        self.zone_button = None
        self.zone_smoke = None
        self.zone_gas = None
        self.zone_lyric_contact = None
        self.zone_lyric_motion = None
        self.zone_lyric_police = None
        self.zone_lyric_temp = None
        self.zone_lyric_keypad = None
        self.zone_lyric_local_alarm = None

    def tests_normal(self):
        """Normal zone."""
        zone = tcz(ZS_NORMAL, None)
        assert zone.partition == "1"
        assert zone.is_bypassed() is False
        assert zone.is_faulted() is False
        assert zone.is_tampered() is False
        assert zone.is_low_battery() is False
        assert zone.is_troubled() is False
        assert zone.is_triggered() is False

    def tests_update(self):
        """Test updates to a zone."""
        self.assertFalse(self.zone_normal.is_faulted())
        self.zone_normal._update(ZONE_FAULTED)
        self.assertTrue(self.zone_normal.is_faulted())
        self.zone_normal._update(ZS_NORMAL)
        self.assertFalse(self.zone_normal.is_faulted())

        self.assertFalse(self.zone_normal.is_low_battery())
        self.zone_normal._update(ZONE_LOW_BATTERY)
        self.assertTrue(self.zone_normal.is_low_battery())
        self.zone_normal._update(ZS_NORMAL)
        self.assertFalse(self.zone_normal.is_low_battery())

    def tests_update_wrong_zone(self):
        """Test updates to the wrong zone."""
        zone_temp = ZS_NORMAL.copy()
        zone_temp["ZoneID"] = "99"
        with pytest.raises(Exception):
            assert self.zone_normal._update(zone_temp)

    def tests_bypassed(self):
        """Bypassed zone."""
        self.assertTrue(self.zone_bypassed.is_bypassed())
        self.assertFalse(self.zone_bypassed.is_faulted())
        self.assertFalse(self.zone_bypassed.is_tampered())
        self.assertFalse(self.zone_bypassed.is_low_battery())
        self.assertFalse(self.zone_bypassed.is_troubled())
        self.assertFalse(self.zone_bypassed.is_triggered())

    def tests_bypass(self):
        """Bypass a zone."""
        self.assertFalse(self.zone_normal.is_bypassed())
        self.zone_normal._mark_as_bypassed()
        self.assertTrue(self.zone_normal.is_bypassed())

    def tests_faulted(self):
        """Faulted zone."""
        self.assertFalse(self.zone_faulted.is_bypassed())
        self.assertTrue(self.zone_faulted.is_faulted())
        self.assertFalse(self.zone_faulted.is_tampered())
        self.assertFalse(self.zone_faulted.is_low_battery())
        self.assertFalse(self.zone_faulted.is_troubled())
        self.assertFalse(self.zone_faulted.is_triggered())

    def tests_tampered(self):
        """Faulted zone."""
        self.assertFalse(self.zone_tampered.is_bypassed())
        self.assertFalse(self.zone_tampered.is_faulted())
        self.assertTrue(self.zone_tampered.is_tampered())
        self.assertFalse(self.zone_tampered.is_low_battery())
        self.assertTrue(self.zone_tampered.is_troubled())
        self.assertFalse(self.zone_tampered.is_triggered())

    def tests_low_battery(self):
        """Zone with low battery."""
        self.assertFalse(self.zone_low_battery.is_bypassed())
        self.assertFalse(self.zone_low_battery.is_faulted())
        self.assertFalse(self.zone_low_battery.is_tampered())
        self.assertTrue(self.zone_low_battery.is_low_battery())
        self.assertFalse(self.zone_low_battery.is_troubled())
        self.assertFalse(self.zone_low_battery.is_triggered())

    def tests_bypassed_low_battery(self):
        """Zone bypassed with low battery."""
        self.assertTrue(self.zone_bypassed_low_battery.is_bypassed())
        self.assertFalse(self.zone_bypassed_low_battery.is_faulted())
        self.assertFalse(self.zone_bypassed_low_battery.is_tampered())
        self.assertTrue(self.zone_bypassed_low_battery.is_low_battery())
        self.assertFalse(self.zone_bypassed_low_battery.is_troubled())
        self.assertFalse(self.zone_bypassed_low_battery.is_triggered())

    def tests_trouble_low_battery(self):
        """Zone with low battery and trouble."""
        self.assertFalse(self.zone_trouble_low_battery.is_bypassed())
        self.assertFalse(self.zone_trouble_low_battery.is_faulted())
        self.assertTrue(self.zone_trouble_low_battery.is_tampered())
        self.assertTrue(self.zone_trouble_low_battery.is_low_battery())
        self.assertTrue(self.zone_trouble_low_battery.is_troubled())
        self.assertFalse(self.zone_trouble_low_battery.is_triggered())

    def tests_triggered(self):
        """Normal zone."""
        self.assertFalse(self.zone_triggered.is_bypassed())
        self.assertFalse(self.zone_triggered.is_faulted())
        self.assertFalse(self.zone_triggered.is_tampered())
        self.assertFalse(self.zone_triggered.is_low_battery())
        self.assertFalse(self.zone_triggered.is_troubled())
        self.assertTrue(self.zone_triggered.is_triggered())

    def tests_type_button(self):
        """Button zone."""
        self.assertFalse(self.zone_normal.is_type_button())
        self.assertTrue(self.zone_button.is_type_button())
        self.assertFalse(self.zone_smoke.is_type_button())
        self.assertFalse(self.zone_gas.is_type_button())
        self.assertTrue(self.zone_lyric_police.is_type_button())
        self.assertFalse(self.zone_lyric_contact.is_type_button())
        self.assertFalse(self.zone_lyric_local_alarm.is_type_button())
        self.assertFalse(self.zone_lyric_motion.is_type_button())
        self.assertFalse(self.zone_lyric_temp.is_type_button())
        self.assertFalse(self.zone_lyric_keypad.is_type_button())

    def tests_type_security(self):
        """Security zone."""
        self.assertTrue(self.zone_normal.is_type_security())
        self.assertTrue(self.zone_button.is_type_security())
        self.assertFalse(self.zone_smoke.is_type_security())
        self.assertFalse(self.zone_gas.is_type_security())
        self.assertTrue(self.zone_lyric_contact.is_type_security())
        self.assertTrue(self.zone_lyric_local_alarm.is_type_security())
        self.assertTrue(self.zone_lyric_motion.is_type_security())
        self.assertTrue(self.zone_lyric_police.is_type_security())
        self.assertFalse(self.zone_lyric_temp.is_type_security())
        self.assertFalse(self.zone_lyric_keypad.is_type_security())

    def tests_type_fire(self):
        """Fire zone."""
        self.assertFalse(self.zone_normal.is_type_fire())
        self.assertFalse(self.zone_button.is_type_fire())
        self.assertTrue(self.zone_smoke.is_type_fire())
        self.assertFalse(self.zone_gas.is_type_fire())
        self.assertFalse(self.zone_lyric_contact.is_type_fire())
        self.assertFalse(self.zone_lyric_local_alarm.is_type_fire())
        self.assertFalse(self.zone_lyric_motion.is_type_fire())
        self.assertFalse(self.zone_lyric_police.is_type_fire())
        self.assertFalse(self.zone_lyric_temp.is_type_fire())

    def tests_type_carbon_monoxide(self):
        """Carbon monoxide zone."""
        self.assertFalse(self.zone_normal.is_type_carbon_monoxide())
        self.assertFalse(self.zone_button.is_type_carbon_monoxide())
        self.assertFalse(self.zone_smoke.is_type_carbon_monoxide())
        self.assertTrue(self.zone_gas.is_type_carbon_monoxide())
        self.assertFalse(self.zone_lyric_contact.is_type_carbon_monoxide())
        self.assertFalse(self.zone_lyric_local_alarm.is_type_carbon_monoxide())
        self.assertFalse(self.zone_lyric_motion.is_type_carbon_monoxide())
        self.assertFalse(self.zone_lyric_police.is_type_carbon_monoxide())
        self.assertFalse(self.zone_lyric_temp.is_type_carbon_monoxide())

    def tests_type_motion(self):
        """Motion zone."""
        self.assertFalse(self.zone_normal.is_type_motion())
        self.assertFalse(self.zone_button.is_type_motion())
        self.assertFalse(self.zone_smoke.is_type_motion())
        self.assertFalse(self.zone_gas.is_type_motion())
        self.assertFalse(self.zone_lyric_contact.is_type_motion())
        self.assertFalse(self.zone_lyric_local_alarm.is_type_motion())
        self.assertTrue(self.zone_lyric_motion.is_type_motion())
        self.assertFalse(self.zone_lyric_police.is_type_motion())
        self.assertFalse(self.zone_lyric_temp.is_type_motion())

    def tests_type_keypad(self):
        """Keypad test."""
        self.assertTrue(self.zone_lyric_keypad.is_type_keypad())
        self.assertFalse(self.zone_lyric_temp.is_type_keypad())

    def tests_type_temperature(self):
        """Temperature test."""
        self.assertFalse(self.zone_lyric_keypad.is_type_temperature())
        self.assertTrue(self.zone_lyric_temp.is_type_temperature())


def test_proa7_zones():
    """Test ProA7."""

    zone_medical = {
        "ZoneDescription": "Gas",
        "PartitionId": "1",
        "ZoneTypeId": ZoneType.PROA7_MEDICAL,
        "CanBeBypassed": 0,
        "ZoneStatus": ZoneStatus.TAMPER,
    }

    zone = tcz(zone_medical, None)
    assert zone.is_type_medical() is True
    assert zone.is_type_button() is True
    assert zone.is_tampered() is True


def test_unknown_type():
    """Test unknown ZoneType."""
    zone_unknown = {
        "ZoneDescription": "Unknown",
        "PartitionId": "1",
        "ZoneTypeId": 12345,
        "CanBeBypassed": 0,
        "ZoneStatus": ZoneStatus.NORMAL,
    }

    zone = tcz(zone_unknown, None)
    assert zone.zone_type_id == 12345
    assert zone._unknown_type_reported is True


def test_unknown_status():
    """Test unknown ZoneStatus."""
    zone_unknown = {
        "ZoneDescription": "Unknown",
        "PartitionId": "1",
        "ZoneTypeId": 12345,
        "CanBeBypassed": 0,
    }

    # invalid status (i.e. None or a string) should raise exception
    with pytest.raises(TotalConnectError):
        tcz(zone_unknown, None)

    # unknown but valid status provided, should not raise
    zone_unknown["ZoneStatus"] = 255
    zone = tcz(zone_unknown, None)
    assert zone.status == 255

def test_bypass():
    """Test bypassing a zone."""
    location = Mock()
    zone_data = {
        "ZoneDescription": "MyZone",
        "PartitionId": "1",
        "ZoneTypeId": ZoneType.SECURITY,
        "CanBeBypassed": 0,
        "ZoneStatus": ZoneStatus.NORMAL,
    }

    # should do nothing if zone cannot be bypassed
    zone = tcz(zone_data, location)
    zone.bypass()
    location.zone_bypass.assert_not_called()

    # now make zone bypassable
    zone_data["CanBeBypassed"] = 1
    zone = tcz(zone_data, location)
    zone.bypass()
    location.zone_bypass.assert_called_once()



"""Test TotalConnectDevice."""

from copy import deepcopy
from unittest.mock import Mock

import pytest
from const import PARTITION_DETAILS_1, PARTITION_DISARMED

from total_connect_client.client import ArmingHelper
from total_connect_client.const import ArmingState
from total_connect_client.exceptions import PartialResponseError, TotalConnectError
from total_connect_client.partition import TotalConnectPartition


def tests_partition():
    """Test __init__()."""
    test_partition = TotalConnectPartition(PARTITION_DETAILS_1, None)
    assert test_partition.partitionid == 1

    test_partition._update(PARTITION_DISARMED)
    assert test_partition.arming_state.value == PARTITION_DISARMED["ArmingState"]

    unknown = deepcopy(PARTITION_DISARMED)
    unknown["ArmingState"] = "999999999"
    with pytest.raises(TotalConnectError):
        test_partition._update(unknown)

    # remove ArmingState
    data = deepcopy(PARTITION_DETAILS_1)
    del data["ArmingState"]
    with pytest.raises(PartialResponseError):
        test_partition._update(data)

    # no info
    with pytest.raises(PartialResponseError):
        test_partition._update(None)


def tests_arm_disarm():
    """Test arm and disarm functions."""
    location = Mock()
    partition = TotalConnectPartition(PARTITION_DETAILS_1, location)

    location.arm_away.return_value = None
    ArmingHelper(partition).arm_away()

    location.arm_stay.return_value = None
    ArmingHelper(partition).arm_stay()

    location.arm_stay_instant.return_value = None
    ArmingHelper(partition).arm_stay_instant()

    location.arm_away_instant.return_value = None
    ArmingHelper(partition).arm_away_instant()

    location.arm_stay_night.return_value = None
    ArmingHelper(partition).arm_stay_night()

    location.disarm.return_value = None
    ArmingHelper(partition).disarm()


def tests_arming_state():
    """Test status functions."""

    partition = TotalConnectPartition(PARTITION_DETAILS_1, None)

    assert partition.arming_state.is_disarming() is False
    assert partition.arming_state.is_disarmed() is True
    assert partition.arming_state.is_arming() is False
    assert partition.arming_state.is_armed() is False
    assert partition.arming_state.is_armed_away() is False
    assert partition.arming_state.is_armed_custom_bypass() is False
    assert partition.arming_state.is_armed_home() is False
    assert partition.arming_state.is_armed_night() is False
    assert partition.arming_state.is_pending() is False
    assert partition.arming_state.is_triggered_police() is False
    assert partition.arming_state.is_triggered_fire() is False
    assert partition.arming_state.is_triggered_gas() is False
    assert partition.arming_state.is_triggered() is False


def tests_proa7_states():
    """Test ProA7 status functions."""

    partition_info = {
        "PartitionID": 1,
        "PartitionName": "Test1",
    }

    armed_stay = [
        ArmingState.ARMED_STAY_PROA7,
        ArmingState.ARMED_STAY_BYPASS_PROA7,
        ArmingState.ARMED_STAY_INSTANT_PROA7,
        ArmingState.ARMED_STAY_INSTANT_BYPASS_PROA7
    ]

    armed_night = [
        ArmingState.ARMED_STAY_NIGHT_BYPASS_PROA7,
        ArmingState.ARMED_STAY_NIGHT_INSTANT_PROA7,
        ArmingState.ARMED_STAY_NIGHT_INSTANT_BYPASS_PROA7,
    ]

    for state in armed_stay:
        partition_info["ArmingState"] = state
        partition = TotalConnectPartition(partition_info, None)
        assert partition.arming_state.is_armed() is True
        assert partition.arming_state.is_armed_home() is True
        assert partition.arming_state.is_armed_night() is False
        assert partition.arming_state.is_armed_away() is False

    for state in armed_night:
        partition_info["ArmingState"] = state
        partition = TotalConnectPartition(partition_info, None)
        assert partition.arming_state.is_armed() is True
        assert partition.arming_state.is_armed_home() is True
        assert partition.arming_state.is_armed_night() is True
        assert partition.arming_state.is_armed_away() is False



"""
total_connect_client package

TotalConnectClient is the main class for the Total Connect interface.

Its accessors may return instances of TotalConnectLocation, TotalConnectPartition,
and TotalConnectZone (from .location, .partition, and .zone respectively), but
users of this interface never create those themselves.
"""

from . import client, const, zone

TotalConnectClient = client.TotalConnectClient
ArmingHelper = client.ArmingHelper

ZoneStatus = zone.ZoneStatus
ZoneType = zone.ZoneType
ArmingState = const.ArmingState
ArmType = const.ArmType
ResultCode = const._ResultCode

__all__ = [
    "TotalConnectClient",
    "ArmType",
    "ArmingState",
    "ArmingHelper",
    "ZoneType",
    "ZoneStatus",
]



"""Test REST API from the command line."""

import getpass
import logging
import sys
import time

from total_connect_client.client import TotalConnectClient
from total_connect_client.const import ArmType
from total_connect_client.exceptions import TotalConnectError

logging.basicConfig(filename="test.log", level=logging.DEBUG)
LOGGER = logging.getLogger(__name__)

if len(sys.argv) != 3:
    print("usage:  username location1=usercode1,location2=usercode2 \n")
    sys.exit()

USERNAME = sys.argv[1]
USERCODES = dict(x.split("=") for x in sys.argv[2].split(","))
PASSWORD = getpass.getpass()

print("Starting test script")
LOGGER.debug("Starting test script")
TC = TotalConnectClient(USERNAME, PASSWORD, USERCODES)

print("Data from client startup\n\n")
print(TC)

print("Attempt to sync.")
LOGGER.debug("Attempt to sync")
for location_id in TC.locations:
    try:
        TC.locations[location_id].sync_panel()
    except TotalConnectError as error:
        LOGGER.error(error)
        print(error)

print("Attempt to validate usercodes.")
LOGGER.debug("Attempt validate usercodes")
for location_id in USERCODES:
    result = TC.locations[int(location_id)].validate_usercode(USERCODES[location_id])
    print(
        f"Usercode {USERCODES[location_id]} valid at location {location_id}: {result}"
    )

print("Attempt zone bypass all.")
LOGGER.debug("Attempt zone bypass all")
for location_id in TC.locations:
    try:
        TC.locations[location_id].zone_bypass_all()
    except TotalConnectError as error:
        LOGGER.error(error)
        print(error)
"""
print("Attempt to arm stay.")
LOGGER.debug("Attempt arm stay")
for location_id in TC.locations:
    try:
        TC.locations[location_id].arm(ArmType.STAY)
    except TotalConnectError as error:
        LOGGER.error(error)
        print(error)
"""
for location_id in TC.locations:
    try:
        TC.locations[location_id].get_panel_meta_data()
    except TotalConnectError as error:
        LOGGER.error(error)
        print(error)

print("Sleep 5 seconds....")
time.sleep(5)
print("Attempt to disarm.")
LOGGER.debug("Attempt disarm")
for location_id in TC.locations:
    try:
        TC.locations[location_id].disarm()
    except TotalConnectError as error:
        LOGGER.error(error)
        print(error)


print("Attempt to clear bypass.")
LOGGER.debug("Attempt clear bypass")
for location_id in TC.locations:
    try:
        TC.locations[location_id].clear_bypass()
    except TotalConnectError as error:
        LOGGER.error(error)
        print(error)

print(TC.times_as_string())

LOGGER.debug("Attempt to log out")
TC.log_out()


sys.exit()



"""Sync system from the command line."""

import getpass
import logging
import sys
import time

from total_connect_client.client import TotalConnectClient

logging.basicConfig(filename="test.log", level=logging.DEBUG)

if len(sys.argv) != 3:
    print("usage:  username location1=usercode1,location2=usercode2 \n")
    sys.exit()

USERNAME = sys.argv[1]
USERCODES = dict(x.split("=") for x in sys.argv[2].split(","))
PASSWORD = getpass.getpass()

TC = TotalConnectClient(USERNAME, PASSWORD, USERCODES)

print("Attempt to bypass all zones.")
for location_id in TC.locations:
    print(f"Bypassing all for location {location_id}")
    TC.locations[location_id].zone_bypass_all()

print("All zones bypassed.  Please verify zones are bypassed as expected.")
print("Bypass will be cleared in 2 minutes.")
time.sleep(60)
print("Bypass will be cleared in 1 minute.")
time.sleep(60)

print("Clearing bypasses.")
for location_id in TC.locations:
    print(f"Clearing bypassing all for location {location_id}")
    TC.locations[location_id].clear_bypass()

print("Cleared zone bypasses.  Please verify zones are faulted/open as expected.")

print("\n\nThanks for helping test.  Please report results on Github.\n\n")

sys.exit()



"""Total Connect Device."""

from typing import Any, Dict


class TotalConnectDevice:
    """Device class for Total Connect."""

    def __init__(self, info: Dict[str, Any]) -> None:
        """Initialize device based on DeviceInfoBasic."""
        self.deviceid = info.get("DeviceID")
        self.name = info.get("DeviceName")
        self.class_id = info.get("DeviceClassID")
        self.serial_number = info.get("DeviceSerialNumber")
        self.security_panel_type_id = info.get("SecurityPanelTypeID")
        self.serial_text = info.get("DeviceSerialText")
        self._doorbell_info: Dict[str, Any] = {}
        self._video_info: Dict[str, Any] = {}
        self._unicorn_info: Dict[str, Any] = {}

        flags = info.get("DeviceFlags")
        if not flags:
            self.flags = {}
        else:
            self.flags = dict(x.split("=") for x in flags.split(","))

    def __str__(self) -> str:
        """Return a string that is printable."""
        data = (
            f"DEVICE {self.deviceid} - {self.name}\n"
            f"  ClassID: {self.class_id}\n"
            f"  Security Panel Type ID: {self.security_panel_type_id}\n"
            f"  Serial Number: {self.serial_number}\n"
            f"  Serial Text: {self.serial_text}\n"
        )

        data = data + "  Device Flags:\n"
        for key, value in self.flags.items():
            data = data + f"    {key}: {value}\n"

        data = data + "  WifiDoorbellInfo:\n"
        for key, value in self._doorbell_info.items():
            data = data + f"    {key}: {value}\n"

        data = data + "  VideoPIRInfo:\n"
        for key, value in self._video_info.items():
            data = data + f"    {key}: {value}\n"

        data = data + "  UnicornInfo:\n"
        for key, value in self._unicorn_info.items():
            data = data + f"    {key}: {value}\n"

        return data

    @property
    def doorbell_info(self) -> Dict[str, Any]:
        """Return doorbell info."""
        return self._doorbell_info

    @doorbell_info.setter
    def doorbell_info(self, data: Dict[str, Any]) -> None:
        """Set values based on WifiDoorBellInfo object."""
        if data:
            self._doorbell_info = data

    def is_doorbell(self) -> bool:
        """Return true if a doorbell."""
        if self._doorbell_info and self._doorbell_info["IsExistingDoorBellUser"] == 1:
            return True

        if (
            self._unicorn_info
            and self._unicorn_info["DeviceVariant"] == "home.dv.doorbell"
        ):
            return True

        return False

    @property
    def video_info(self) -> Dict[str, Any]:
        """VideoPIR info."""
        return self._video_info

    @video_info.setter
    def video_info(self, data: Dict[str, Any]) -> None:
        """Set values based on VideoPIRInfo object."""
        if data:
            self._video_info = data

    @property
    def unicorn_info(self) -> Dict[str, Any]:
        """Unicorn info."""
        return self._video_info

    @unicorn_info.setter
    def unicorn_info(self, data: Dict[str, Any]) -> None:
        """Set values based on UnicornInfo object."""
        if data:
            self._unicorn_info = data



"""Test your system from the command line."""

import getpass
import logging
import sys

from total_connect_client.client import TotalConnectClient

logging.basicConfig(filename="test.log", level=logging.DEBUG)

if len(sys.argv) < 2 or len(sys.argv) > 3:
    print("usage:  python3 -m total_connect_client username [password]\n")
    sys.exit(1)


USERNAME = sys.argv[1]
if len(sys.argv) == 2:
    PASSWORD = getpass.getpass()
else:
    PASSWORD = sys.argv[2]

TC = TotalConnectClient(USERNAME, PASSWORD)

print(TC)

TC.log_out()



"""Test your system from the command line."""

import logging
import sys
from pprint import pprint

from total_connect_client.client import TotalConnectClient

logging.basicConfig(filename="test.log", level=logging.DEBUG)

if len(sys.argv) != 3:
    print("usage:  username password\n")
    sys.exit()


USERNAME = sys.argv[1]
PASSWORD = sys.argv[2]

TC = TotalConnectClient(USERNAME, PASSWORD)

location_id = next(iter(TC.locations))

print("auto_bypass_low_battery: " + str(TC.auto_bypass_low_battery))

print("\n\n\n")

print("--- panel meta data ---")
meta_data = TC.get_panel_meta_data(location_id)
pprint(meta_data)

print("\n\n\n")

print(TC.locations[location_id])

for z in TC.locations[location_id].zones:
    print(TC.locations[location_id].zones[z])



"""Test your system from the command line."""

import getpass
import logging
import sys

from ..client import TotalConnectClient
from ..const import ArmType

logging.basicConfig(filename="test.log", level=logging.DEBUG)


def usage():
    """Print usage instructions."""
    print("usage:  username location1=usercode1,location2=usercode2 arm_type \n")
    print("arm types:")
    for i in list(ArmType):
        print(f"  {i.name}: {i.value}")
    sys.exit()


if len(sys.argv) != 4:
    usage()

USERNAME = sys.argv[1]
USERCODES = dict(x.split("=") for x in sys.argv[2].split(","))
arm_type = ArmType(int(sys.argv[3]))
PASSWORD = getpass.getpass()

TC = TotalConnectClient(USERNAME, PASSWORD, USERCODES)

for location_id in TC.locations:
    TC.locations[location_id].arm(arm_type)

print(f"Armed: {arm_type}")

sys.exit()



"""Test TotalConnectDevice."""

from const import DEVICE_INFO_BASIC_1

from total_connect_client.device import TotalConnectDevice


def tests_init():
    """Test __init__()."""
    test_device = TotalConnectDevice(DEVICE_INFO_BASIC_1)
    assert test_device.deviceid == DEVICE_INFO_BASIC_1["DeviceID"]

    # test with missing flags
    del DEVICE_INFO_BASIC_1["DeviceFlags"]
    test_device = TotalConnectDevice(DEVICE_INFO_BASIC_1)
    assert not test_device.flags



"""Test your system from the command line."""

import getpass
import logging
import sys

from total_connect_client.client import TotalConnectClient

logging.basicConfig(filename="test.log", level=logging.DEBUG)

if len(sys.argv) != 3:
    print("usage:  username location1=usercode1,location2=usercode2 \n")
    sys.exit()

USERNAME = sys.argv[1]
USERCODES = dict(x.split("=") for x in sys.argv[2].split(","))
PASSWORD = getpass.getpass()

TC = TotalConnectClient(USERNAME, PASSWORD, USERCODES)

for location_id in TC.locations:
    TC.locations[location_id].disarm()

print("Disarmed.")

sys.exit()



"""Tests TotalConnectLocation."""

from unittest.mock import Mock 
from total_connect_client.location import TotalConnectLocation

from const import LOCATION_INFO_BASIC_NORMAL

def tests_location_basic():
    """Tests basic location info."""
    location = TotalConnectLocation(LOCATION_INFO_BASIC_NORMAL, Mock())
    location.location_id = 123456
